import functools
import inspect
from hashlib import md5
from json import dumps
from typing import AnyStr, List
from uuid import UUID

from ruleau.constants import OverrideLevel
from ruleau.docs import description, parameters, title
from ruleau.exceptions import TopLevelRuleRequiresNameException


class Rule:
    def __init__(
        self,
        func,
        name: AnyStr,
        depends_on: List["Rule"],
        override_level: OverrideLevel,
        lazy_dependencies: bool,
    ):
        """
        :param func: User defined rule
        :param name: User defined human readable name of the rule
        :param depends_on: Rule dependencies
        :param override_level: Override level
        :param lazy_dependencies: Flag to switch loading of rule dependencies lazily
        """
        self.name = name
        # Validate the rule, make sure the name is always set for a rule
        self.validate()
        # Set the user defined function
        self.func = func
        self.depends_on = depends_on
        self.override_level = override_level
        self.__name__ = func.__name__
        self.lazy_dependencies = lazy_dependencies
        # Get parsed data, set a unique rule ID
        rule_hash = md5(dumps(self.parse()).encode())
        self.id = str(UUID(rule_hash.hexdigest()))

        # This preserves the original Docstring on the decorated function
        # which allows DocTest to detect the function
        functools.update_wrapper(self, func)

    def __str__(self) -> str:
        return self.__name__

    def __call__(self, *args, **kwargs) -> bool:
        return self.func(*args, **kwargs)

    def validate(self):
        """Validator to check if top level rule has a human readable name
        :raises: TopLevelRuleRequiresNameException
        """
        if not self.name or not isinstance(self.name, str):
            raise TopLevelRuleRequiresNameException()

    def parse(self):
        return {
            "function_name": self.__name__,
            "name": self.name,
            "title": title(self.__name__),
            "override_level_name": self.override_level.name,
            "override_level": self.override_level.value,
            "source": inspect.getsource(self.func),
            "docstring": self.func.__doc__,
            "description": description(self.func.__doc__),
            "parameters": parameters(self.func.__doc__),
            "dependencies": [dependent.parse() for dependent in self.depends_on],
        }



import json
import logging
from typing import Any, AnyStr, Dict, List, Optional
from urllib.parse import urljoin

import requests

from ruleau.decorators import api_request
from ruleau.rule import Rule

logger = logging.getLogger(__name__)


class ApiAdapter:

    base_url: AnyStr
    base_path: AnyStr
    api_key: Optional[AnyStr]

    def __init__(
        self,
        base_url: AnyStr,
        api_key: Optional[AnyStr] = None,
    ):
        """
        :param base_url: Base URL of the ruleau API
        :param api_key: (Optional) API key to authenticate with the API
        """
        self.base_url = base_url
        self.base_path = "/api/v1/"
        self.api_key = api_key

    @api_request
    def sync_case(self, client_case_id, rule_id, payload) -> dict:
        """
        Synchronise case with API
        :param rule_id:
        :param client_case_id:
        :param payload:
        :return:
        """
        response = requests.get(
            urljoin(self.base_url, f"{self.base_path}cases/{client_case_id}")
        )
        if response.status_code == 200:
            response = requests.patch(
                urljoin(self.base_url, f"{self.base_path}cases/{client_case_id}"),
                json={
                    "client_case_id": client_case_id,
                    "payload": payload,
                    "rule": rule_id,
                    "status": "OPEN",
                },
            )
            if response.status_code != 200:
                raise Exception(f"Failed to update case: {response.text}")
        else:
            response = requests.post(
                urljoin(self.base_url, f"{self.base_path}cases"),
                json={
                    "client_case_id": client_case_id,
                    "payload": payload,
                    "rule": rule_id,
                    "status": "OPEN",
                },
            )
            if response.status_code != 201:
                raise Exception(f"Failed to create case: {response.text}")

        return response.json()

    @api_request
    def sync_rule(self, rule: Rule, order: int, dependencies: List[AnyStr]) -> dict:
        """
        Synchronise rule with API
        :param dependencies:
        :param order:
        :param rule:
        :return:
        """
        # Check if the rule exists
        response = requests.get(
            urljoin(self.base_url, f"{self.base_path}rules/{rule.id}"),
        )
        # If the rule doesn't exist, create it
        if response.status_code == 404:
            parsed_rule = rule.parse()
            # NOTE: This is probably a future requirement,
            # which might change slightly later
            rule_name = parsed_rule["name"] or parsed_rule["function_name"]
            response = requests.post(
                urljoin(self.base_url, f"{self.base_path}rules"),
                json={
                    "id": rule.id,
                    "name": rule_name,
                    "description": parsed_rule["description"],
                    "override_level": parsed_rule["override_level"],
                    "order": order,
                    "dependencies": dependencies,
                    "result": {},
                },
            )
            if response.status_code != 201:
                raise Exception(f'Unable to save rule {rule_name}": {response.text}')
        return response.json()

    @api_request
    def create_result(
        self,
        client_case_id: AnyStr,
        rule: Rule,
    ) -> dict:
        """
        Update rule result after it's execution
        :param client_case_id:
        :param rule:
        :return:
        """
        # Update the rule results
        response = requests.post(
            urljoin(
                self.base_url,
                f"{self.base_path}cases/{client_case_id}/rules/{rule.id}/results",
            )
        )
        if response.status_code != 201:
            raise Exception(
                f"Failed to create result {client_case_id}@{rule.id}: {response.text}"
            )
        return response.json()

    @api_request
    def update_result(
        self,
        client_case_id: AnyStr,
        rule: Rule,
        api_result: Dict[AnyStr, Any],
        execution_result: "ExecutionResult",  # noqa: F821
    ) -> dict:
        """
        Update rule result after it's execution
        :param client_case_id:
        :param rule:
        :param api_result:
        :param execution_result:
        :return:
        """
        # Update the rule results
        response = requests.patch(
            urljoin(
                self.base_url,
                f"{self.base_path}cases/{client_case_id}/"
                f"rules/{rule.id}/results/{api_result['id']}",
            ),
            json={
                "result": execution_result.result,
                "payload": execution_result.payload.accessed,
                "override": execution_result.override,
                "original_result": execution_result.original_result,
            },
        )
        if response.status_code != 200:
            raise Exception(
                f"Failed to store rule result "
                f"{client_case_id}@{rule.id}: {response.text}"
            )
        return response.json()

    @api_request
    def fetch_override(
        self, client_case_id: AnyStr, rule: Rule
    ) -> Optional[Dict[AnyStr, Any]]:
        """
        Fetch rule overrides
        :param rule:
        :param client_case_id: UUID that identifies a previously established case
        :return: a ruleau overrides Optional[Dict[AnyStr, Any]]
        """
        override = {}
        parsed_rule = rule.parse()
        response = requests.get(
            urljoin(
                self.base_url,
                f"{self.base_path}cases/{client_case_id}/overrides/search",
            ),
            params={"rule_name": parsed_rule["name"] or parsed_rule["function_name"]},
        )
        if response.status_code == 200:
            override = response.json()
        return override



class TopLevelRuleRequiresNameException(Exception):
    """Exception raised if top level aggregators don't have a human readable name"""

    pass


class MethodNotAllowedException(Exception):
    """Exception raised if a forbidden RuleauDict method is called"""

    pass


class CaseIdRequiredException(Exception):
    """Exception raised if a json path for case identifier is not found"""

    pass


class CannotOverrideException(Exception):
    """Exception raised if a API tries to override a rule marked as NO_OVERRIDE"""


class DuplicateRuleNameException(Exception):
    """Exception raised if more than 1 rule has same name"""

    pass



import argparse
import importlib
import re
import sys
from argparse import Namespace
from os import mkdir, path
from typing import AnyStr

from jinja2 import Template


def trim(docstring):
    """trim function implementation from PEP-257
    https://www.python.org/dev/peps/pep-0257/#id18

    :param docstring: Docstring for a function
    """
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:  # pragma: no cover
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    if "\n" in docstring:
        trimmed.append("")
    # Return a single string:
    return "\n".join(trimmed)


def description(docstring: AnyStr) -> AnyStr:
    """Takes a rule docstring, and returns just the lines
    representing a description

    :param docstring: Docstring for a function
    """
    result = trim(docstring or "")
    if "\n\n" in result:
        return result[: result.index("\n\n")]
    return result


def parameters(docstring: AnyStr):
    """Parse list of parameters out of the docstring

    :param docstring: The docstring of the rule
    :returns: [{"name": ..., "description": ...}, ...]
    """
    PARAM_REGEX = re.compile(
        ":(?P<name>[\*\s\w]+): (?P<description>.*?)\Z"  # noqa: W605
    )

    params = {}
    if docstring:
        lines = docstring.split("\n")
        for line in lines:
            match = PARAM_REGEX.findall(line.strip())
            if match:
                param_name, param_description = match[0]
                params.update({param_name: param_description})

    return params


def title(name: AnyStr) -> AnyStr:
    """Convert the function name into a title

    :param name: Name of a function
    """
    return name.replace("_", " ").title()


def generate_documentation(rules) -> AnyStr:
    """Returns a HTML string, documenting the passed in rule and its dependants

    :param rules: List of rules to generate document of
    """

    with open(
        path.join(path.dirname(path.realpath(__file__)), "html", "documentation.html")
    ) as f:
        doc_template = Template(f.read())

    enriched_rules = [rule.parse() for rule in rules]
    return doc_template.render(rules=enriched_rules)


def render_doc_for_module(module_file) -> AnyStr:
    """Render the documentation from a rule module
    :param module_file:
    :return:
    """
    spec = importlib.util.spec_from_file_location(module_file, module_file)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)

    # Pick rules to document
    rules = [
        rule
        for rule in vars(config).values()
        if getattr(type(rule), "__name__", None) == "Rule"
    ]
    if not rules:
        raise Exception("No rules found in {}.".format(module_file))

    return generate_documentation(rules)


def generate_and_save_to_file(input_files, output_dir) -> int:
    """Generate the documentation and save to a file in provided output dir
    :param input_files: Rule modules provided by user via CLI
    :param output_dir: Output directory of generated HTML file
    :return:
    """
    generated_docs = []

    if len(input_files) == 0:
        raise Exception("No file(s) supplied to generate documentation for.")

    if not path.exists(output_dir):
        mkdir(output_dir)

    for target_file in input_files:
        if not path.exists(target_file):
            raise ValueError(f"{target_file} does not exist")

        documentation = render_doc_for_module(target_file)

        output_filename = path.join(
            output_dir, "{}.html".format(path.basename(target_file).split(".")[0])
        )

        with open(output_filename, "w") as f:
            f.write(documentation)
        generated_docs.append(output_filename)

    print(
        f"{len(generated_docs)} doc{'s' if len(generated_docs) > 1 else ''} generated."
    )
    print("\n".join(generated_docs))

    return 0


def get_arguments(args) -> Namespace:
    """Parses arguments for the ruleau-docs command
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default="./",
        help="Directory for generated documents",
    )
    parser.add_argument("files", nargs="*")
    return parser.parse_args(args)


def main() -> int:  # pragma: no cover
    """Console script for deft document generation.
    USAGE:
    ```bash
    $ ruleau-docs [--output-dir=<argument>] filename ...
    ```
    """
    args = get_arguments(sys.argv[1:])
    return generate_and_save_to_file(args.files, args.output_dir)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover
